#include "X3D_FileValidatorGeometry3DVisitor.h"
#include "X3D_IndexedFaceSet.h"
#include "X3D_Coordinate.h"
#include "X3D_Normal.h"
#include "X3D_Color.h"
#include "X3D_ColorRGBA.h"
#include "X3D_TextureCoordinate.h"

#include <iostream>

using namespace std;

namespace X3DTK {
namespace X3D {

FileValidatorGeometry3DVisitor::FileValidatorGeometry3DVisitor()
: Geometry3DVisitor()
{
  stateVariables = GraphTraversal::getInstanceOf<FileValidatorStateVariables>();
  
  // Enter functions.
  defineEnterFunction(&FileValidatorGeometry3DVisitor::enterIndexedFaceSet);
}

FileValidatorGeometry3DVisitor::~FileValidatorGeometry3DVisitor()
{
}

void FileValidatorGeometry3DVisitor::enterIndexedFaceSet(IndexedFaceSet *I) const
{
  cx3d << "node " << stateVariables->getNode() << " " << I->getTypeName() << endl;
  
  ////////////////// arrays //////////////////////////////////////////////////////////
  
  int pointArraySize = 0;
  int normalArraySize = 0;
  int colorArraySize = 0;
  int texCoordArraySize = 0;
  
  X3D::Coordinate *CoordinateNode = dynamic_cast<X3D::Coordinate *>(I->getCoord());
  if (CoordinateNode != 0)
    pointArraySize = CoordinateNode->getPoint().size();
  
  X3D::Normal *NormalNode = dynamic_cast<X3D::Normal *>(I->getNormal());
  if (NormalNode != 0)
    normalArraySize = NormalNode->getVector().size();
  
  if (I->getColor() != 0)
  {
    if (I->getColor()->getTypeName() == "ColorRGBA")
      colorArraySize = dynamic_cast<X3D::ColorRGBA *>(I->getColor())->getColor().size();
    else  
      colorArraySize = dynamic_cast<X3D::Color *>(I->getColor())->getColor().size();
  }
  
  X3D::TextureCoordinate *TexCoordNode = dynamic_cast<X3D::TextureCoordinate *>(I->getTexCoord());
  if (TexCoordNode != 0)
    texCoordArraySize = TexCoordNode->getPoint().size();
  
  bool hasPoint = (pointArraySize != 0);
  bool hasNormal = (normalArraySize != 0);
  bool hasColor = (colorArraySize != 0);
  bool hasTexCoord = (texCoordArraySize != 0);

  ////////////////// indexes /////////////////////////////////////////////////////////

  MFInt32 coordIndex;
  MFInt32 normalIndex;
  MFInt32 colorIndex;
  MFInt32 texCoordIndex;
 
  coordIndex = I->getCoordIndex();
  colorIndex = I->getColorIndex();
  normalIndex = I->getNormalIndex();
  texCoordIndex = I->getTexCoordIndex();

  bool hasColorIndex = !colorIndex.empty();
  bool hasNormalIndex = !normalIndex.empty();
  bool hasTexCoordIndex = !texCoordIndex.empty();
  
  ////////////////// iterators ///////////////////////////////////////////////////////
    
  MFInt32::const_iterator itCoordIndex = coordIndex.begin();
  MFInt32::const_iterator itColorIndex;
  MFInt32::const_iterator itNormalIndex;
  MFInt32::const_iterator itTexCoordIndex;
  unsigned int itFaceIndex = 0; 
    
  if (hasColorIndex)
    itColorIndex = colorIndex.begin();
  else
    itColorIndex = coordIndex.begin();
  
  if (hasNormalIndex)
    itNormalIndex = normalIndex.begin();  
  else
    itNormalIndex = coordIndex.begin();
  
  if (hasTexCoordIndex)
    itTexCoordIndex = texCoordIndex.begin();
  else
    itTexCoordIndex = coordIndex.begin();
 
  
  ////////////////// loop ////////////////////////////////////////////////////////////
  
  SFBool normalPerVertex = I->getNormalPerVertex();
  SFBool colorPerVertex = I->getColorPerVertex();
        
  while (itCoordIndex != coordIndex.end())
  {
    while ((*itCoordIndex != -1) && (itCoordIndex != coordIndex.end()))
    {
      if (hasPoint)
        if ((*itCoordIndex >= pointArraySize) || (*itCoordIndex < 0))
        {
          cx3d << "  '->coordinate " << *itCoordIndex << " doesn't exist!" << endl;
          stateVariables->setValidity(false);
        }
      
      ////////////// normals /////////////////////////////////////////////////////////
        
      if (hasNormal)
        if (itNormalIndex == normalIndex.end())
        {
          cx3d << "  '->normalIndex too small!" << endl;
          hasNormal = false;
          stateVariables->setValidity(false);
        }
        else
        {
          if (hasNormalIndex)
          {
            if ((*itNormalIndex >= normalArraySize) || (*itNormalIndex < 0))
            {
              cx3d << "  '->normal of index " << *itNormalIndex << " doesn't exist!" << endl;
              stateVariables->setValidity(false);
            }
          }  
          else
          {
            if (static_cast<unsigned int>(itFaceIndex) >= normalArraySize)
            {
              cx3d << "  '->normal of face " << itFaceIndex << " doesn't exist!" << endl;
              stateVariables->setValidity(false);
            }  
          }  
        }  

      if (hasColor) 
        if (itColorIndex == colorIndex.end())
        {
          cx3d << "  '->colorIndex too small!" << endl;
          hasColor = false;
          stateVariables->setValidity(false);
        }
        else
        {
          if (hasColorIndex)
          {
            if ((*itColorIndex >= colorArraySize) || (*itColorIndex < 0))
            {
              cx3d << "  '->color of index " << *itColorIndex << " doesn't exist!" << endl;
              stateVariables->setValidity(false);
            }
          }  
          else
          {
            if (static_cast<unsigned int>(itFaceIndex) >= colorArraySize)
            {
              cx3d << "  '->color of face " << itFaceIndex << " doesn't exist!" << endl;
              stateVariables->setValidity(false);
            }  
          }  
        }   

      if (hasTexCoord) 
        if (itTexCoordIndex == texCoordIndex.end())
        {
          cx3d << "  '->texCoordIndex too small!" << endl;
          hasTexCoord = false;
          stateVariables->setValidity(false);
        }
        else 
        {
          if ((*itTexCoordIndex >= texCoordArraySize) || (*itTexCoordIndex < 0))
          {
            cx3d << "  '->texCoord of index " << *itTexCoordIndex << " doesn't exist!" << endl;
            stateVariables->setValidity(false);
          }
        }  
        
      if (hasPoint) 
        ++itCoordIndex;
      if (hasColor && colorPerVertex)
        ++itColorIndex;     
      if (hasNormal && normalPerVertex)
        ++itNormalIndex;      
      if (hasTexCoord)   
        ++itTexCoordIndex;
    }
    
    if (itNormalIndex == normalIndex.end())
    {
      cx3d << "  '->normalIndex too small!" << endl;
      hasNormal = false;
      stateVariables->setValidity(false);
    }
    
    if (hasColor) 
      if (itColorIndex == colorIndex.end())
      {
        cx3d << " '->colorIndex too small!" << endl;
        hasColor = false;
        stateVariables->setValidity(false);
      }

    if (hasTexCoord) 
      if (itTexCoordIndex == texCoordIndex.end())
      {
        cx3d << "  '->texCoordIndex too small!" << endl;
        hasTexCoord = false;
        stateVariables->setValidity(false);
      }

    if (hasPoint) 
      ++itCoordIndex;
    if (hasColor)
      ++itColorIndex;     
    if (hasNormal)
      ++itNormalIndex;      
    if (hasTexCoord)   
      ++itTexCoordIndex;
    ++itFaceIndex;  
  }

  stateVariables->addNode();  
  
  /*Coordinate *CoordNode = static_cast<Coordinate *>(I->getCoord());
  if (CoordNode == NULL)
    return;
   
  int pointArraySize = 0;
  int normalArraySize = 0;
  int colorArraySize = 0;
  int texCoordArraySize = 0;
  
  X3D::Coordinate *CoordinateNode = dynamic_cast<X3D::Coordinate *>(I->getCoord());
  if (CoordinateNode != 0)
    pointArraySize = CoordNode->getPoint().size();
  

  X3D::Normal *NormalNode = dynamic_cast<X3D::Normal *>(I->getNormal());
  if (NormalNode != 0)
    normalArraySize = NormalNode->getVector().size();
  
  if (I->getColor() != 0)
  {
    if (I->getColor()->getTypeName() == "ColorRGBA")
      colorArraySize = dynamic_cast<X3D::ColorRGBA *>(I->getColor())->getColor().size();
    else  
      colorArraySize = dynamic_cast<X3D::Color *>(I->getColor())->getColor().size();
  }
  
  X3D::TextureCoordinate *TexCoordNode = dynamic_cast<X3D::TextureCoordinate *>(I->getTexCoord());
  if (TexCoordNode != 0)
    texCoordArraySize = TexCoordNode->getPoint().size();
  
  const MFInt32 &coordIndex = I->getCoordIndex();
  const MFInt32 &colorIndex = I->getColorIndex();
  const MFInt32 &normalIndex = I->getNormalIndex();
  const MFInt32 &texCoordIndex = I->getTexCoordIndex();

  bool hasPoint = (I->getCoord() != 0);
  bool hasNormal = (I->getNormal() != 0);
  bool hasColor = (I->getColor() != 0);
  bool hasTexCoord = (I->getTexCoord() != 0);
  
  SFBool normalPerVertex = I->getNormalPerVertex();
  SFBool colorPerVertex = I->getColorPerVertex();
  
  MFInt32::const_iterator itCoordIndex = coordIndex.begin();
  MFInt32::const_iterator itColorIndex;
  MFInt32::const_iterator itNormalIndex;
  MFInt32::const_iterator itTexCoordIndex;
  
  if (colorIndex.empty())
    itColorIndex = coordIndex.begin();
  else
    itColorIndex = colorIndex.begin();  
  
  if (normalIndex.empty())
    itNormalIndex = coordIndex.begin();
  else
    itNormalIndex = normalIndex.begin();  
  
  if (texCoordIndex.empty())
    itTexCoordIndex = coordIndex.begin();
  else
    itTexCoordIndex = texCoordIndex.begin();
    
  while (itCoordIndex != coordIndex.end())
  {
    while ((*itCoordIndex != -1) 
        && (itCoordIndex != coordIndex.end()))
    {
      if ((hasPoint) 
       && ((*itCoordIndex >= pointArraySize) || (*itCoordIndex < 0)))
      {
        cx3d << "point " << *itCoordIndex << " doesn't exist!" << endl;
        stateVariables->setValidity(false);
      }

      if (hasNormal)
        if (itNormalIndex == normalIndex.end())
        {
          cx3d << "normalIndex too small!" << endl;
          hasNormal = false;
          stateVariables->setValidity(false);
        }
        else
          if ((*itNormalIndex >= normalArraySize) || (*itNormalIndex < 0))
          {
            cx3d << "normal " << *itNormalIndex << " doesn't exist!" << endl;
            stateVariables->setValidity(false);
          }

      if (hasColor) 
        if (itColorIndex == colorIndex.end())
        {
          cx3d << "colorIndex too small!" << endl;
          hasColor = false;
          stateVariables->setValidity(false);
        }
        else
          if ((*itColorIndex >= colorArraySize) || (*itColorIndex < 0))
          {
            cx3d << "color " << *itColorIndex << " doesn't exist!" << endl;
            stateVariables->setValidity(false);
          }

      if (hasTexCoord) 
        if (itTexCoordIndex == texCoordIndex.end())
        {
          cx3d << "texCoordIndex too small!" << endl;
          hasTexCoord = false;
          stateVariables->setValidity(false);
        }
        else 
          if ((*itTexCoordIndex >= texCoordArraySize) || (*itTexCoordIndex < 0))
          {
            cx3d << "texCoord " << *itTexCoordIndex << " doesn't exist!" << endl;
            stateVariables->setValidity(false);
          }
        
      if (hasPoint) 
        ++itCoordIndex;
      if (hasColor && colorPerVertex)
        ++itColorIndex;     
      if (hasNormal && normalPerVertex)
        ++itNormalIndex;      
      if (hasTexCoord)   
        ++itTexCoordIndex;
    }
    
    if (itNormalIndex == normalIndex.end())
    {
      cx3d << "normalIndex too small!" << endl;
      hasNormal = false;
      stateVariables->setValidity(false);
    }
    if (hasColor) 
      if (itColorIndex == colorIndex.end())
      {
        cx3d << "colorIndex too small!" << endl;
        hasColor = false;
        stateVariables->setValidity(false);
      }

    if (hasTexCoord) 
      if (itTexCoordIndex == texCoordIndex.end())
      {
        cx3d << "texCoordIndex too small!" << endl;
        hasTexCoord = false;
        stateVariables->setValidity(false);
      }

    if (hasPoint) 
      ++itCoordIndex;
    if (hasColor)
      ++itColorIndex;     
    if (hasNormal)
      ++itNormalIndex;      
    if (hasTexCoord)   
      ++itTexCoordIndex;
  }*/
  
  
  
}

}
}
